////////////////////////////////////////////////////////////////////////////
//
// K1999.cpp
//
// car driver for TORCS
// created:	(c) Remi Coulom March 2000
//
// modified:	2008 Andrew Sumner andrewsumner@yahoo.com
// modified:	2009 John Isham isham.john@gmail.com
// last update:	2014 Michel Luc mluc@cern91.net
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
////////////////////////////////////////////////////////////////////////////

#include <math.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>

#include "tgf.h" 
#include "track.h" 
#include "car.h"
#include "raceman.h" 
#include "robot.h" 
#include "robottools.h"

#include "raceline.h"
#include "line.h"

////////////////////////////////////////////////////////////////////////////
// Parameters
////////////////////////////////////////////////////////////////////////////

//
// These parameters are for the computation of the path
//
static const int Iterations = 100;     // Number of smoothing operations
//static const double DivLength = 3.0;   // Length of path elements in meters
 
static const double SecurityR = 100.0; // Security radius
static double SideDistExt = 2.0; // Security distance wrt outside
static double SideDistInt = 1.0; // Security distance wrt inside          

/////////////////////////////////////////////////////////////////////////////
// Some utility macros and functions
/////////////////////////////////////////////////////////////////////////////

static double Mag(double x, double y)
{
  return sqrt((x * x) + (y * y));
}
 
static double Min(double x1, double x2)
{
 if (x1 < x2)
  return x1;
 else
  return x2;
}
 
static double Max(double x1, double x2)
{
 if (x1 < x2)
  return x2;
 else
  return x1;
}
 
LRaceLine::~LRaceLine()
{
#if 0
  delete [] tSegDist;
  delete [] tSegIndex;
  delete [] tElemLength;
  delete [] tx;
  delete [] ty;
  delete [] tLane;
  delete [] tRInverse;
  delete [] tSpeed;
  delete [] tDistance;
  delete [] tMaxSpeed;
  delete [] txLeft;
  delete [] tyLeft;
  delete [] txRight;
  delete [] tyRight;
  delete [] tFriction;

  delete [] tzLeft;
  delete [] tzRight;

  delete [] turnSpeedSegStart;
  delete [] turnSpeedSegEnd;
  delete [] segTurnSpeed;

  delete [] MspeedSegStart;
  delete [] MspeedSegEnd;
  delete [] segMaxSpeed;
#endif

}

/////////////////////////////////////////////////////////////////////////////
// Update tx and ty arrays
/////////////////////////////////////////////////////////////////////////////
void LRaceLine::UpdateTxTy(int i, int rl)
{
 tx[rl][i] = tLane[rl][i] * txRight[i] + (1 - tLane[rl][i]) * txLeft[i];
 ty[rl][i] = tLane[rl][i] * tyRight[i] + (1 - tLane[rl][i]) * tyLeft[i];
}                                                                               

/////////////////////////////////////////////////////////////////////////////
// Set segment info
/////////////////////////////////////////////////////////////////////////////
void LRaceLine::SetSegmentInfo(const tTrackSeg *pseg, double d, int i, double l)
{
 if (pseg)
 {
  tSegDist[pseg->id] = d;
  tSegIndex[pseg->id] = i;
  tElemLength[pseg->id] = l;
  if (pseg->id >= Segs)
   Segs = pseg->id + 1;
 }
}

/////////////////////////////////////////////////////////////////////////////
// Split the track into small elements
// ??? constant width supposed
/////////////////////////////////////////////////////////////////////////////
void LRaceLine::SplitTrack(tTrack *ptrack, int rl)
{
 // DivLength = GfParmGetNum( car->_carHandle, BT_SECT_PRIV, "DivLength", (char *)NULL, 3 );
 DivLength = 3;
 Segs = 0;
 
 const tTrackSeg *psegCurrent = ptrack->seg;

 double Distance = 0;
 double Angle = psegCurrent->angle[TR_ZS];
 double xlPos = psegCurrent->vertex[TR_SL].x;
 double ylPos = psegCurrent->vertex[TR_SL].y;
 double xrPos = psegCurrent->vertex[TR_SR].x;
 double yrPos = psegCurrent->vertex[TR_SR].y;
 double xPos  = (xlPos + xrPos) / 2;
 double yPos  = (ylPos + yrPos) / 2;

 int i = 0;

 do
 {
  int Divisions = 1 + int(psegCurrent->length / DivLength);
  double Step = psegCurrent->length / Divisions;

  SetSegmentInfo(psegCurrent, Distance + Step, i, Step);
  for (int j = Divisions; --j >= 0;)
  {
   double cosine = cos(Angle);
   double sine = sin(Angle);
   
   if (psegCurrent->type == TR_STR)
   {
    xPos += cosine * Step;
    yPos += sine * Step;
   }
   else
   {
    double r = psegCurrent->radius;
    double Theta = psegCurrent->arc / Divisions;
    double L = 2 * r * sin(Theta / 2);
    double x = L * cos(Theta / 2);
    double y;
    if (psegCurrent->type == TR_LFT)
    {
     Angle += Theta;
     y = L * sin(Theta / 2);
    }
    else
    {
     Angle -= Theta;
     y = -L * sin(Theta / 2);
    }
    xPos += x * cosine - y * sine;
    yPos += x * sine + y * cosine;
   }

   double dx = -psegCurrent->width * sin(Angle) / 2;
   double dy = psegCurrent->width * cos(Angle) / 2;
   txLeft[i] = xPos + dx;
   tyLeft[i] = yPos + dy;
   txRight[i] = xPos - dx;
   tyRight[i] = yPos - dy;
   tLane[rl][i] = 0.5;
   tFriction[i] = psegCurrent->surface->kFriction;
   if (tFriction[i] < 1) // ??? ugly trick for dirt
   {
    // fprintf("%s: dirt hack segment:%d\n",__FUNCTION__,i); 

    //tFriction[i] *= 0.90;
    fDirt = 1;
    SideDistInt = -1.5;
    SideDistExt = 0.0;
   }
   UpdateTxTy(i, rl);

   Distance += Step;
   tDistance[i] = Distance;
   i++;
  }

  psegCurrent = psegCurrent->next;
 }
 while (psegCurrent != ptrack->seg);

 Divs = i - 1;

 Width = psegCurrent->width;
 Length = Distance;

}
 
/////////////////////////////////////////////////////////////////////////////
// Compute the inverse of the radius
/////////////////////////////////////////////////////////////////////////////
double LRaceLine::GetRInverse(int prev, double x, double y, int next, int rl)
{
 double x1 = tx[rl][next] - x;
 double y1 = ty[rl][next] - y;
 double x2 = tx[rl][prev] - x;
 double y2 = ty[rl][prev] - y;
 double x3 = tx[rl][next] - tx[rl][prev];
 double y3 = ty[rl][next] - ty[rl][prev];
 
 double det = x1 * y2 - x2 * y1;
 double n1 = x1 * x1 + y1 * y1;
 double n2 = x2 * x2 + y2 * y2;
 double n3 = x3 * x3 + y3 * y3;
 double nnn = sqrt(n1 * n2 * n3);
 
 return 2 * det / nnn;
}

/////////////////////////////////////////////////////////////////////////////
// Change lane value to reach a given radius
/////////////////////////////////////////////////////////////////////////////
void LRaceLine::AdjustRadius(int prev, int i, int next, double TargetRInverse, int rl, double Security)
{
 double OldLane = tLane[rl][i];
 
 //
 // Start by aligning points for a reasonable initial lane
 //
 tLane[rl][i] = (-(ty[rl][next] - ty[rl][prev]) * (txLeft[i] - tx[rl][prev]) +
		 (tx[rl][next] - tx[rl][prev]) * (tyLeft[i] - ty[rl][prev])) /
                ( (ty[rl][next] - ty[rl][prev]) * (txRight[i] - txLeft[i]) -
		  (tx[rl][next] - tx[rl][prev]) * (tyRight[i] - tyLeft[i]));
 if (tLane[rl][i] < -0.2)
  tLane[rl][i] = -0.2;
 else if (tLane[rl][i] > 1.2)
  tLane[rl][i] = 1.2;
 UpdateTxTy(i, rl);
 
 //
 // Newton-like resolution method
 //
 const double dLane = 0.0001;
 
 double dx = dLane * (txRight[i] - txLeft[i]);
 double dy = dLane * (tyRight[i] - tyLeft[i]);
 
 double dRInverse = GetRInverse(prev, tx[rl][i] + dx, ty[rl][i] + dy, next, rl);
 
 if (dRInverse > 0.000000001)
 {
  tLane[rl][i] += (dLane / dRInverse) * TargetRInverse;
 
/****************** Get Extern/Intern Margin **********************/
#if 0
	ExtMargin = GfParmGetNum( car->_carHandle, BT_SECT_PRIV, "extmargin", (char *)NULL, 2.0 );
	IntMargin = GfParmGetNum( car->_carHandle, BT_SECT_PRIV, "extmargin", (char *)NULL, 1.0 );

	double ExtLane = ExtMargin + Security) / Width;
  	double IntLane = IntMargin + Security) / Width;
#endif

  double ExtLane = (getExtMargin(i) + Security) / Width;
  double IntLane = (getIntMargin(i) + Security) / Width;
  if (ExtLane > 0.5)
   ExtLane = 0.5;
  if (IntLane > 0.5)
   IntLane = 0.5;

  if (TargetRInverse >= 0.0)
  {
   if (tLane[rl][i] < IntLane)
    tLane[rl][i] = IntLane;
   if (1 - tLane[rl][i] < ExtLane)
   {
    if (1 - OldLane < ExtLane)
     tLane[rl][i] = Min(OldLane, tLane[rl][i]);
    else
     tLane[rl][i] = 1 - ExtLane;
   }
  }
  else
  {
   if (tLane[rl][i] < ExtLane)
   {
    if (OldLane < ExtLane)
     tLane[rl][i] = Max(OldLane, tLane[rl][i]);
    else
     tLane[rl][i] = ExtLane;
   }
   if (1 - tLane[rl][i] < IntLane)
    tLane[rl][i] = 1 - IntLane;
  }
 }
 
 UpdateTxTy(i, rl);
}

/////////////////////////////////////////////////////////////////////////////
// Smooth path
/////////////////////////////////////////////////////////////////////////////
void LRaceLine::Smooth(int Step, int rl)
{
 int prev = ((Divs - Step) / Step) * Step;
 int prevprev = prev - Step;
 int next = Step;
 int nextnext = next + Step;
 
 for (int i = 0; i <= Divs - Step; i += Step)
 {
  double ri0 = GetRInverse(prevprev, tx[rl][prev], ty[rl][prev], i, rl);
  double ri1 = GetRInverse(i, tx[rl][next], ty[rl][next], nextnext, rl);
  double lPrev = Mag(tx[rl][i] - tx[rl][prev], ty[rl][i] - ty[rl][prev]);
  double lNext = Mag(tx[rl][i] - tx[rl][next], ty[rl][i] - ty[rl][next]);

  double TargetRInverse = (lNext * ri0 + lPrev * ri1) / (lNext + lPrev);
 
  double Security = lPrev * lNext / (8 * SecurityR);

  if (rl == LINE_RL)
  {
   if (ri0 * ri1 > 0)
   {
    double ac1 = fabs(ri0);
    double ac2 = fabs(ri1);
    {
     if (ac1 < ac2)
      ri0 += 0.12 * (ri1 - ri0);
     else if (ac2 < ac1)
      ri1 += 0.12 * (ri0 - ri1);
    }
 
    TargetRInverse = (lNext * ri0 + lPrev * ri1) / (lNext + lPrev);
   }
	}

  AdjustRadius(prev, i, next, TargetRInverse, rl, Security);
 
  prevprev = prev;
  prev = i;
  next = nextnext;
  nextnext = next + Step;
  if (nextnext > Divs - Step)
   nextnext = 0;
 }
}

/////////////////////////////////////////////////////////////////////////////
// Interpolate between two control points
/////////////////////////////////////////////////////////////////////////////
void LRaceLine::StepInterpolate(int iMin, int iMax, int Step, int rl)
{
 int next = (iMax + Step) % Divs;
 if (next > Divs - Step)
  next = 0;
 
 int prev = (((Divs + iMin - Step) % Divs) / Step) * Step;
 if (prev > Divs - Step)
  prev -= Step;
 
 double ir0 = GetRInverse(prev, tx[rl][iMin], ty[rl][iMin], iMax % Divs, rl);
 double ir1 = GetRInverse(iMin, tx[rl][iMax % Divs], ty[rl][iMax % Divs], next, rl);
 for (int k = iMax; --k > iMin;)
 {
  double x = double(k - iMin) / double(iMax - iMin);
  double TargetRInverse = x * ir1 + (1 - x) * ir0;
  AdjustRadius(iMin, k, iMax % Divs, TargetRInverse, rl);
 }
}
 
/////////////////////////////////////////////////////////////////////////////
// Calls to StepInterpolate for the full path
/////////////////////////////////////////////////////////////////////////////
void LRaceLine::Interpolate(int Step, int rl)
{
 if (Step > 1)
 {
  int i;
  for (i = Step; i <= Divs - Step; i += Step)
   StepInterpolate(i - Step, i, Step, rl);
  StepInterpolate(i - Step, Divs, Step, rl);
 }
}


/*===========================================================*/
/*                  SET DATA                                 */
/*===========================================================*/

#define BT_SECT_PRIV (char *)"tux private"
#define myModuleName (char *)"tuxy_dtm"

void LRaceLine::setRwData( tTrack* t, void **carParmHandle, tSituation *s)
{
  float attvalue;
  int i;
  //char attname[15];

  line_verbose = 0;
  steer_verbose = 0;
  LineIndex =0;
  LineSave = 0;

  if ( (NULL == carParmHandle) || (NULL == *carParmHandle) )
    {
      printf("%s: NULL cfg handle, aborting\n", myModuleName);
      return;
    }
  
  if (0.0 != (attvalue = GfParmGetNum(*carParmHandle, BT_SECT_PRIV, (char *)"lineverbose", (char*) NULL, 0.0)))
    line_verbose = (int)attvalue;
  if (0.0 != (attvalue = GfParmGetNum(*carParmHandle, BT_SECT_PRIV, (char *)"steerverbose", (char*) NULL, 0.0)))
    steer_verbose = (int)attvalue;
  if (0.0 != (attvalue = GfParmGetNum(*carParmHandle, BT_SECT_PRIV, (char *)"lineindex", (char*) NULL, 0.0)))
    LineIndex = (int)attvalue;
  if (0.0 != (attvalue = GfParmGetNum(*carParmHandle, BT_SECT_PRIV, (char *)"linesave", (char*) NULL, 0.0)))
    LineSave = (int)attvalue;
#if 0
    if ( s->_raceType != RM_TYPE_PRACTICE )
     {
   	LineIndex = 0;
	LineSave = 0;
      }
#endif

/* Load data (corner speed, max speed & margin) from data file */
    char filename[256];
    sprintf(filename, "%sdrivers/%s/data/%s.data", GetDataDir(), myModuleName, t->internalname);
	if ( s->_raceType == RM_TYPE_QUALIF ) {
   	sprintf(filename, "%sdrivers/%s/data/%s.qualif", GetDataDir(), myModuleName, t->internalname);
      	}

  std::string line;
  //std::ifstream tfile(filename, ios::in);
  std::ifstream tfile(filename);
 if (!tfile.is_open()) {
	std::cout << ">>>> setLineData() Error: no data file found!" << std::endl;
	for ( i=0; i<MAX_TDATA; i++)
    	{
		turnSpeedSegStart[i] = -1.0;
      		turnSpeedSegEnd[i] = 0.0;
      		segTurnSpeed[i] = 20.0;

		brakeDistSegStart[i] = -1.0;
		brakeDistSegEnd[i] = 0.0;
		segBrakeDist[i] = 19.0;
		
		MspeedSegStart[i] = -1.0;
		MspeedSegEnd[i] = 0.0;
		segMaxSpeed[i] = 2000.0;

		extMargSegStart[i] = -1.0;
		extMargSegEnd[i] = 0.0;
		segExtMargin[i] = 1.4;

		intMargSegStart[i] = -1.0;
		intMargSegEnd[i] = 0.0;
		segIntMargin[i] = 0.3;
	}

 }else {
	std::cout << ">>>> setLineData() OK read " << filename << std::endl;
	for ( i=0; i<MAX_TDATA; i++)
    	{
		getline (tfile, line);
      		turnSpeedSegStart[i] = (int)atof(line.c_str());
		getline (tfile, line);
      		turnSpeedSegEnd[i] = (int)atof(line.c_str());
		getline (tfile, line);
      		segTurnSpeed[i] = atof(line.c_str());
		getline (tfile, line);

		getline (tfile, line);
      		brakeDistSegStart[i] = (int)atof(line.c_str());
		getline (tfile, line);
      		brakeDistSegEnd[i] = (int)atof(line.c_str());
		getline (tfile, line);
      		segBrakeDist[i] = atof(line.c_str());
		getline (tfile, line);

		getline (tfile, line);
      		MspeedSegStart[i] = (int)atof(line.c_str());
		getline (tfile, line);
      		MspeedSegEnd[i] = (int)atof(line.c_str());
		getline (tfile, line);
      		segMaxSpeed[i] = (5.0/18.0) * atof(line.c_str());
		getline (tfile, line);

		getline (tfile, line);
      		extMargSegStart[i] = (int)atof(line.c_str());
		getline (tfile, line);
      		extMargSegEnd[i] = (int)atof(line.c_str());
		getline (tfile, line);
      		segExtMargin[i] = atof(line.c_str());
		getline (tfile, line);

		getline (tfile, line);
      		intMargSegStart[i] = (int)atof(line.c_str());
		getline (tfile, line);
      		intMargSegEnd[i] = (int)atof(line.c_str());
		getline (tfile, line);
      		segIntMargin[i] = atof(line.c_str());
		getline (tfile, line);
	}

 }
 tfile.close();


   /* Save data in another format */
  if ( LineSave )
 {
 char filename[256];
    sprintf(filename, "%sdrivers/%s/tracks/%s.data", GetLocalDir(), myModuleName, t->internalname);
  std::ofstream tfile;
  tfile.open (filename);
  if (!tfile.is_open())
	return;
  for (int i = 0; i <MAX_TDATA; i++) {
    if (i > 0) {
      tfile << std::endl;
    }
    tfile << turnSpeedSegStart[i] << std::endl;
    tfile << turnSpeedSegEnd[i] << std::endl;
    tfile << segTurnSpeed[i] << std::endl;
    tfile << std::endl;
    tfile << MspeedSegStart[i] << std::endl;
    tfile << MspeedSegEnd[i] << std::endl;
    tfile << segMaxSpeed[i] << std::endl;
    tfile << std::endl;
    tfile << extMargSegStart[i] << std::endl;
    tfile << extMargSegEnd[i] << std::endl;
    tfile << segExtMargin[i] << std::endl;
    tfile << std::endl;
    tfile << intMargSegStart[i] << std::endl;
    tfile << intMargSegEnd[i] << std::endl;
    tfile << segIntMargin[i] << std::endl;
    tfile << std::endl;
  }
  tfile.close();

   }

  return;
}

/**********************************************************/
double LRaceLine::getCornerSpeed(int Div)
{
  int i;

  for ( i=0; i<MAX_TDATA; i++)
    {
      if ( ( -1 != turnSpeedSegStart[i] ) &&
	   (Div >= turnSpeedSegStart[i] ) &&
	   (Div <= turnSpeedSegEnd[i])    )
	return (segTurnSpeed[i]);

    }
  //return turnSpeed;
  return CornerSpeed;

}

/**********************************************************/
double LRaceLine::getBrakeDist(int Div)
{
  int i;

  for ( i=0; i<MAX_TDATA; i++)
    {
      if ( ( -1 != brakeDistSegStart[i] ) &&
	   (Div >= brakeDistSegStart[i] ) &&
	   (Div <= brakeDistSegEnd[i])    )
	return (segBrakeDist[i]);

    }
  return BrakeDist;

}

/**********************************************************/
double LRaceLine::getIntMargin(int Div)
{
  int i;
  IntMargin = 0.3;
  for ( i=0; i<MAX_TDATA; i++)
    {
      if ( ( -1 != intMargSegStart[i] ) &&
	   (Div >= intMargSegStart[i] ) &&
	   (Div <= intMargSegEnd[i])    )
	return (segIntMargin[i]);

    }

  return IntMargin;
}

/**********************************************************/
double LRaceLine::getExtMargin(int Div)
{
  int i;
  ExtMargin = 1.4;
  for ( i=0; i<MAX_TDATA; i++)
    {
      if ( ( -1 != extMargSegStart[i] ) &&
	   (Div >= extMargSegStart[i] ) &&
	   (Div <= extMargSegEnd[i])    )
	return (segExtMargin[i]);

    }

  return ExtMargin;
}

/*===========================================================*/
/*                INIT TRACK                                 */
/*===========================================================*/
void LRaceLine::InitTrack(tTrack* track, tSituation *p)
{

	//double turnSpeed = GfParmGetNum( car->_carHandle, BT_SECT_PRIV, "turnspeed", (char *)NULL, 21.0 );
	minTurnInverse = GfParmGetNum( car->_carHandle, BT_SECT_PRIV, "turnreverse", (char *)NULL, 0.002 );
	speedAdjust = GfParmGetNum( car->_carHandle, BT_SECT_PRIV, "speedadjust", (char *)NULL, 2.0 );
	brakeDist = GfParmGetNum( car->_carHandle, BT_SECT_PRIV, "brakedelay", (char *)NULL, 15.0 ); 

 // split track
  memset(tRInverse, 0, sizeof(tRInverse));
  memset(tLane, 0, sizeof(tLane));

 for (int rl=LINE_MID; rl<=LINE_RL; rl++)
 {
	/*Split the track into small elements*/
  	SplitTrack(track, rl);
	/* save the data in a file */
	StoreData(track);

  // Smoothing loop
  int Iter = (rl == LINE_MID ? Iterations/4 : Iterations);
  for (int Step = 128; (Step /= 2) > 0;)
  {
   for (int i = Iter * int(sqrt((float) Step)); --i >= 0;)
    Smooth(Step, rl);
   Interpolate(Step, rl);
  }


  // Compute curvature and speed along the path
  for (int i = Divs; --i >= 0;)
  {
    vec2f tmp;
   //double TireAccel = turnSpeed * tFriction[i];
   double TireAccel = getCornerSpeed(i) * tFriction[i];
   if (rl == LINE_MID)
    TireAccel += speedAdjust;
   int next = (i + 1) % Divs;
   int prev = (i - 1 + Divs) % Divs;
 
   double rInverse = GetRInverse(prev, tx[rl][i], ty[rl][i], next, rl);
   /* rinverse from clothoid curvature??  */
   tRInverse[rl][i] = rInverse;
 
   double MaxSpeed;

   if (fabs(rInverse) > minTurnInverse * 1.01)
    MaxSpeed = sqrt(TireAccel / (fabs(rInverse) - minTurnInverse));
   else
    MaxSpeed = 10000;

   //if (fabs(rInverse) > 0.0087 )
   //printf("rInv:%f MaxSpd:%f\n", rInverse,  MaxSpeed);

/*+++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    /*  TODO: increase or decrease speed depending on
   	 	track camber
	 	track slope
		approaching bumps	    */
/*+++++++++++++++++++++++++++++++++++++++++++++++++++++*/

   tSpeed[rl][i] = tMaxSpeed[i] = MaxSpeed;

   if ( LINE_RL == rl )
     {
       int turnSpecs;
       for ( turnSpecs=0; turnSpecs<MAX_TDATA; turnSpecs++)
	 {
	   if ( ( -1 != MspeedSegStart[turnSpecs] ) && (i >= MspeedSegStart[turnSpecs] ) && (i <= MspeedSegEnd[turnSpecs]) )
	     {
	     if (segMaxSpeed[turnSpecs] < tSpeed[rl][i] )
	       {
		 if (line_verbose )
		   GfOut("OVERRIDE max speed in div %d old:%f new:%f\n", i, tSpeed[rl][i], segMaxSpeed[turnSpecs]);
		
		 tSpeed[rl][i] = tMaxSpeed[i] = segMaxSpeed[turnSpecs];
	       }
	     else
	       {
		 if (line_verbose )
		   GfOut("NO OVERRIDE  max speed in div %d old:%f new:%f\n", i, tSpeed[rl][i], segMaxSpeed[turnSpecs]);
	       }
	     }
	 }
     }

  }

  //
  // Anticipate braking
  //
  for (int j = 32; --j >= 0;)
  for (int i = Divs; --i >= 0;)
  {
   //double TireAccel = turnSpeed * tFriction[i];
   double TireAccel = getCornerSpeed(i) * tFriction[i];
   int prev = (i - 1 + Divs) % Divs;
 
   double dx = tx[rl][i] - tx[rl][prev];
   double dy = ty[rl][i] - ty[rl][prev];
   double dist = Mag(dx, dy);
 
   double Speed = (tSpeed[rl][i] + tSpeed[rl][prev]) / 2;
 
   double LatA = tSpeed[rl][i] * tSpeed[rl][i] *
                 (fabs(tRInverse[rl][prev]) + fabs(tRInverse[rl][i])) / 2;
 
#if 0
   double TanA = TireAccel * TireAccel - LatA * LatA;
   if (TanA < 0.0)
    TanA = 0.0;
   TanA = sqrt(TanA) + minTurnInverse * Speed * Speed;
   if (TanA > brakeDist)
    TanA = brakeDist;
#else
   double TanA = TireAccel * TireAccel + minTurnInverse * Speed * Speed - LatA * LatA;
   //double brakedelay = brakeDist + (rl == LINE_MID ? speedAdjust : 0.0);
   double brakedelay = getBrakeDist(i) + (rl == LINE_MID ? speedAdjust : 0.0);
   if (TanA < 0.0)
    TanA = 0.0;
   if (TanA > brakedelay * tFriction[i])
    TanA = brakedelay * tFriction[i];
#endif
 
   double Time = dist / Speed;
   double MaxSpeed = tSpeed[rl][i] + TanA * Time;
   tSpeed[rl][prev] = Min(MaxSpeed, tMaxSpeed[prev]);
  }                                                                              
 }

} 

////////////////////////////////////////////////////////////////////////////
// New race
////////////////////////////////////////////////////////////////////////////
void LRaceLine::NewRace(tCarElt* newcar, tSituation *s)
{
 car = newcar;

 wheelbase = (car->priv.wheel[FRNT_RGT].relPos.x +
              car->priv.wheel[FRNT_LFT].relPos.x -
              car->priv.wheel[REAR_RGT].relPos.x -
              car->priv.wheel[REAR_LFT].relPos.x) / 2;
 wheeltrack = (car->priv.wheel[FRNT_LFT].relPos.y +
               car->priv.wheel[REAR_LFT].relPos.y -
               car->priv.wheel[FRNT_RGT].relPos.y -
               car->priv.wheel[REAR_RGT].relPos.y) / 2;
} 

////////////////////////////////////////////////////////////////////////////
// Car control
////////////////////////////////////////////////////////////////////////////
void LRaceLine::GetRaceLineData(tSituation *s, v2d *target, float *speed, float *avspeed, float *raceoffset, float *lookahead, float *racesteer)
{
 // 
 // Find index in data arrays
 //
 tTrackSeg *seg = car->_trkPos.seg;
 int SegId = car->_trkPos.seg->id;
 double dist = car->_trkPos.toStart;
 if (dist < 0)
  dist = 0;
 if (car->_trkPos.seg->type != TR_STR)
  dist *= car->_trkPos.seg->radius;
 int Index = tSegIndex[SegId] + int(dist / tElemLength[SegId]);
 This = Index; //sectors(fromstart)

Index = (Index + Divs - 5) % Divs;
/*====================================================*/
 /* Display Index used in the data file */

  if (LineIndex)
	GfOut("index: %d\n",Index);
/*====================================================*/
#if 0
	double turnAccel = GfParmGetNum( car->_carHandle, BT_SECT_PRIV, "turnaccel", (char *)NULL, 1.0 );

 double TimeC = 1.0;
 double TimeA = 0.15 + turnAccel/80;
 double TimeB = s->deltaTime*7;
 if ( TimeA > TimeB * 2)
  {
	TimeC = TimeB;
  }
   else TimeC = TimeA;

 static const double Time = TimeC;
#endif

 int steer_verbose = 0;

 static const double Time = s->deltaTime*7; 	// 0.50 + turnAccel/80;
#if 0
 double X4 = car->_pos_X + car->_speed_X * 0.5 / 2;
 double Y4 = car->_pos_Y + car->_speed_Y * 0.5 / 2;
#endif
 double X = car->_pos_X + car->_speed_X * Time / 2;
 double Y = car->_pos_Y + car->_speed_Y * Time / 2;
 *lookahead = 0.0f;

 int tmp_Index = Index;
 while(1)
 {
  Next = (Index + 1) % Divs;
  double dx = tx[LINE_RL][Next] - car->_pos_X;
  double dy = ty[LINE_RL][Next] - car->_pos_Y;
  *lookahead = sqrt(dx*dx + dy*dy);
  if (*lookahead > 10.0f &&
      (tx[LINE_RL][Next] - tx[LINE_RL][Index]) * (X - tx[LINE_RL][Next]) +
      (ty[LINE_RL][Next] - ty[LINE_RL][Index]) * (Y - ty[LINE_RL][Next]) < 0.1)
   break;
  Index = Next;
 }
 Index = tmp_Index;


 if ((tRInverse[LINE_RL][Next] > 0.0 && car->_trkPos.toMiddle < 0.0) ||
		 (tRInverse[LINE_RL][Next] < 0.0 && car->_trkPos.toMiddle > 0.0))
  *lookahead *= MIN(4.0f, 1.5f + fabs(car->_trkPos.toMiddle*0.3));
 else
  *lookahead *= MAX(0.7f, 1.5f - fabs(car->_trkPos.toMiddle*0.2));

#if 0
  double dx = X4 - car->_pos_X;
  double dy = Y4 - car->_pos_Y;
  *lookahead = sqrt(dx*dx + dy*dy);
#endif

#if 1
 if ((tRInverse[LINE_RL][Next] < 0.0 && car->_trkPos.toMiddle > 0.0) ||
     (tRInverse[LINE_RL][Next] > 0.0 && car->_trkPos.toMiddle < 0.0))
 {
  *lookahead *= (float) MAX(1.0f, MIN(3.6f, 1.0f + (MIN(2.6f, fabs(car->_trkPos.toMiddle) / (seg->width/2)) / 2) * (1.0f + fabs(tRInverse[LINE_RL][Next]) * 65.0f + car->_speed_x/120.0f)));
 }
 else if ((tRInverse[LINE_RL][Next] < 0.0 && car->_trkPos.toRight < 5.0) ||
          (tRInverse[LINE_RL][Next] > 0.0 && car->_trkPos.toLeft < 5.0))
 {
  *lookahead *= MAX(0.8f, MIN(1.0f, 1.0f - fabs(tRInverse[LINE_RL][Next])*200.0 * ((5.0f-MIN(car->_trkPos.toRight, car->_trkPos.toLeft))/5.0f)));
 }
#endif

 target->x = (float) tx[LINE_RL][Next];
 target->y = (float) ty[LINE_RL][Next];

 //
 // Find target speed
 //
 double c0 = (tx[LINE_RL][Next] - tx[LINE_RL][Index]) * (tx[LINE_RL][Next] - X) +
             (ty[LINE_RL][Next] - ty[LINE_RL][Index]) * (ty[LINE_RL][Next] - Y);
 double c1 = (tx[LINE_RL][Next] - tx[LINE_RL][Index]) * (X - tx[LINE_RL][Index]) +
             (ty[LINE_RL][Next] - ty[LINE_RL][Index]) * (Y - ty[LINE_RL][Index]);
 
 {
   double sum = c0 + c1;

   c0 /= sum;
   c1 /= sum;
 }

 
 TargetSpeed = (1 - c0) * tSpeed[LINE_RL][Next] + c0 * tSpeed[LINE_RL][Index];
 *avspeed = (float) MAX(10.0f, tSpeed[LINE_MID][Next]);

 /* WARNING: problems on Corkscrew pit in with TargetSpeed being negative? */
 //*speed = (float) MAX(*avspeed, TargetSpeed);
 //*speed = TargetSpeed;
  *speed = MAX( 10.0f, TargetSpeed);


 if ((tRInverse[LINE_RL][Next] > 0.0 && tLane[LINE_RL][Next] > tLane[LINE_RL][Index] && car->_trkPos.toLeft <= tLane[LINE_RL][Next] * Width + 1.0) ||
     (tRInverse[LINE_RL][Next] < 0.0 && tLane[LINE_RL][Next] < tLane[LINE_RL][Index] && car->_trkPos.toLeft >= tLane[LINE_RL][Next] * Width - 1.0))
 {
  *avspeed = MAX(*speed, *avspeed);
 }
 else if ((tRInverse[LINE_RL][Next] > 0.001 && tLane[LINE_RL][Next] < tLane[LINE_RL][Index] && car->_trkPos.toLeft < tLane[LINE_RL][Next] * Width - 1.0) ||
          (tRInverse[LINE_RL][Next] < -0.001 && tLane[LINE_RL][Next] > tLane[LINE_RL][Index] && car->_trkPos.toLeft > tLane[LINE_RL][Next] * Width + 1.0))
 {
  *avspeed *= (float) MAX(0.7f, 1.0 - fabs(tRInverse[LINE_RL][Next])*100);
 }


 double laneoffset = Width/2 - (tLane[LINE_RL][Next] * Width);
 *raceoffset = (float) laneoffset;

 //
 // Find target curvature (for the inside wheel)
 //
 double TargetCurvature = (1 - c0) * tRInverse[LINE_RL][Next] + c0 * tRInverse[LINE_RL][Index];
 if (fabs(TargetCurvature) > 0.01)
 {
  double r = 1 / TargetCurvature;
  if (r > 0)
   r -= wheeltrack / 2;
  else
   r += wheeltrack / 2;
  TargetCurvature = 1 / r;
 }

 //
 // Steering control
 //
 double Error = 0;
 double VnError = 0;
 double carspeed = Mag(car->_speed_X, car->_speed_Y);
 

 //
 // Ideal value
 //
 double steer = atan(wheelbase * TargetCurvature) / car->_steerLock;

 if ( 0 && steer_verbose )
   GfOut("Steer A:%f",steer);

 //
 // Servo system to stay on the pre-computed path
 //
 {
  double dx = tx[LINE_RL][Next] - tx[LINE_RL][Index];
  double dy = ty[LINE_RL][Next] - ty[LINE_RL][Index];
  Error = (dx * (Y - ty[LINE_RL][Index]) - dy * (X - tx[LINE_RL][Index])) / Mag(dx, dy);
 }  

 int Prev = (Index + Divs - 1) % Divs;
#if 0
 int PrevPrev = (Index + Divs - 5) % Divs;
#endif
 int NextNext = (Next + 1) % Divs;
 double Prevdx = tx[LINE_RL][Next] - tx[LINE_RL][Prev];
 double Prevdy = ty[LINE_RL][Next] - ty[LINE_RL][Prev];
 double Nextdx = tx[LINE_RL][NextNext] - tx[LINE_RL][Index];
 double Nextdy = ty[LINE_RL][NextNext] - ty[LINE_RL][Index];
 double dx = c0 * Prevdx + (1 - c0) * Nextdx;
 double dy = c0 * Prevdy + (1 - c0) * Nextdy;
 double n = Mag(dx, dy);
 dx /= n;
 dy /= n;
 double sError = (dx * car->_speed_Y - dy * car->_speed_X) / (carspeed + 0.01);
 double cError = (dx * car->_speed_X + dy * car->_speed_Y) / (carspeed + 0.01);

 VnError = asin(sError);

 VnError = MAX ( (PI/-2.0), MIN( VnError, (PI/2.0)));

 if (0 && fabs(sError) > 0.9 )
   {
     GfOut("tuxy_dtm: %s:1 sError:%f\n", __FUNCTION__, sError);
     steer_verbose = 1;
   }
 if (0 && (cError < 0.2) )
   {
     GfOut("tuxy_dtm: %s:1 cError:%f\n", __FUNCTION__, cError);
     steer_verbose = 1;
   }

 if (cError < 0)
   {
#if 1
     /* Warning! in Ole-road-1 seems to be a lot of damage over bumps
	and sudden change of direction. */
     VnError = PI - VnError;
     //VnError = (2.0 *PI) - VnError;
     //VnError = -1.0 * VnError;
#endif
   }

 if ( 0 && fabs(VnError) > (0.9 * PI) )
   {
     GfOut("%s: %s:3 VnError(pre-norm):%f\n",car->_name, __FUNCTION__,VnError);
     steer_verbose = 1;
   }

 NORM_PI_PI(VnError);

 if ( 0 && steer_verbose )
   {
     GfOut("steer:4 dx:%f dy:%f n:%f\n",dx, dy, n);
     GfOut("steer:4 Error:%f sError:%f VnError:%f cError:%f pre-atan:%f post-atan:%f\n",
	    Error, sError, VnError, cError, 
	    Error * (300.0 / (carspeed + 300.0) / 15.0),
	    atan(Error * (300.0 / (carspeed + 300.0) / 15.0))
	    );
   }

 // Correcting Steer:
	int C_steering = 0;
	float attvalue = 0.0;
	if ((attvalue = GfParmGetNum( car->_carHandle, BT_SECT_PRIV, "steercorrect", (char *)NULL, 0.0 )) != 0.0)
		C_steering = (int)attvalue;
   double steer_correct;

/* turns are closed but the car goes slower */
 steer_correct = (atan(Error * (300.0 / (carspeed + 300.0)) / 15.0) + VnError) / car->_steerLock;
 //NORM_PI_PI(steer_correct);

 /* Warning! The car goes too fast cornering */
 if (C_steering == 1) {
 	steer_correct = MAX ( (PI/-8.0), MIN( steer_correct, (PI/8.0)));
 } else if (C_steering == 2) {
	 steer_correct = MAX ( (PI/-10.0), MIN( steer_correct, (PI/10.0)));
 }
 if ( 0 && fabs(steer_correct) > (PI))
   {
     GfOut("%s: %s:5 steer_correct:%f sError:%f\n",car->_name, __FUNCTION__, steer_correct, sError);
   }
 steer -= steer_correct;

 
 if ( steer_verbose )
   GfOut("steer: B:%f\n",steer);

 //
 // Steer into the skid
 //
 double vx = car->_speed_X;
 double vy = car->_speed_Y;
 double dirx = cos(car->_yaw);
 double diry = sin(car->_yaw);
 double Skid = (dirx * vy - vx * diry) / (carspeed + 0.1);
 if (Skid > 0.9)
  Skid = 0.9;
 if (Skid < -0.9)
  Skid = -0.9;
 steer += (asin(Skid) / car->_steerLock) * 0.9;

 if ( 0 && steer_verbose )
   GfOut(" C:%f",steer);

 double yr = carspeed * TargetCurvature;
 double diff = car->_yaw_rate - yr;
 steer -= (0.08 * (100 / (carspeed + 100)) * diff) / car->_steerLock;

 if ( 0 && steer_verbose )
   GfOut(" D:%f",steer);

 {
  double trackangle = RtTrackSideTgAngleL(&(car->_trkPos));
  double angle = trackangle - car->_yaw;
  NORM_PI_PI(angle);
  angle = -angle;

  if (fabs(angle) > 1.0)
    {
      if ((angle > 0.0 && steer > 0.0) || (angle < 0.0 && steer < 0.0))
	steer = -steer;
    }

 if ( 0 && steer_verbose )
   GfOut(" F:%f",steer);

  if (fabs(angle) > 1.6)
    {
      // we're facing the wrong way.  Set it to full steer in whatever direction for now ...
      if (steer > 0.0)
	steer = 1.0;
      else
	steer = -1.0;
    }

 if ( 0 && steer_verbose )
   GfOut(" G:%f\n",steer);

 }

 *racesteer = steer;
} 

/**********************************************************/
int LRaceLine::isOnLine(int line)
{
 double lane2left = tLane[line][Next] * Width;

 double pos2middle = fabs(car->_trkPos.toLeft - lane2left);

 if ( steer_verbose )
 	fprintf(stderr,"Lane ToLeft:%.3f, Car ToLeft:%.3f, Car ToMiddle:%.3f\n", lane2left, car->_trkPos.toLeft, car->_trkPos.toMiddle);

 if ( pos2middle < MAX(0.1, 1.0 - (car->_speed_x * (car->_speed_x/10))/600))
  {
 	if ( steer_verbose )
		fprintf(stderr,"%s is OnLine: PosToMiddle: %.3f \n", car->_name, pos2middle);
  	return 1;
  }

 return 0;
}

/**********************************************************/
void LRaceLine::GetPoint( float offset, float lookahead, vec2f *rt )
{
 double lane = (Width/2 - offset) / Width;
#if 0
 int ndiv = (Next + 1 + int((lookahead)/DivLength)) % Divs;
 rt->x = (float) (lane * txRight[ndiv] + (1 - lane) * txLeft[ndiv]);
 rt->y = (float) (lane * tyRight[ndiv] + (1 - lane) * tyLeft[ndiv]);
#endif
 double length = 0.0;
 //double la = (double) lookahead * MAX(0.8, car->_speed_x/TargetSpeed); //0.8;
 double la = (double) lookahead * MIN(1.0, MAX(0.8, car->_speed_x/TargetSpeed)); //0.8;
 vec2f last;
 last.x = (float) (lane * txRight[This] + (1 - lane) * txLeft[This]);
 last.y = (float) (lane * tyRight[This] + (1 - lane) * tyLeft[This]);
 //last.x = car->_pos_X;
 //last.y = car->_pos_Y;
 //int ndiv = (Next + 1 + int((lookahead)/DivLength)) % Divs;
 int ndiv = Next, count = 0;
 while (length < la && count < (int) (la/DivLength))
 {
  rt->x = (float) (lane * txRight[ndiv] + (1 - lane) * txLeft[ndiv]);
  rt->y = (float) (lane * tyRight[ndiv] + (1 - lane) * tyLeft[ndiv]);
  double dx = rt->x - last.x;
  double dy = rt->y - last.y;
  double thislength = sqrt(dx*dx + dy*dy);
	length += thislength;
  ndiv = (ndiv + 1) % Divs;
  count++;
  last.x = rt->x;
  last.y = rt->y;
 }
}

/**********************************************************/
double LRaceLine::correctLimit(int line)
{
 // this returns true if we're approaching a corner & are significantly
 // inside the ideal racing line. The idea is to prevent a sudden outwards
 // movement at a time when we should be looking to turn in.
 
 double nlane2left = tLane[line][Next] * Width;
 if ((tRInverse[line][Next] > 0.001 && car->_trkPos.toLeft < nlane2left - 2.0) ||
     (tRInverse[line][Next] < -0.001 && car->_trkPos.toLeft > nlane2left + 2.0))
   //return MAX(0.2, MIN(1.0, 1.0 - fabs(tRInverse[line][Next]) * 80.0));
   return MAX(0.2, MIN(1.0, 1.0 - fabs(tRInverse[line][Next]) * 100.0));

 int nnext = (Next + (int) (car->_speed_x/3)) % Divs;
 double nnlane2left = tLane[line][nnext] * Width;
 if ((tRInverse[line][nnext] > 0.001 && car->_trkPos.toLeft < nnlane2left - 2.0) ||
     (tRInverse[line][nnext] < -0.001 && car->_trkPos.toLeft > nnlane2left + 2.0))
   //return MAX(0.2, MIN(1.0, 1.0 - fabs(tRInverse[line][nnext]) * 40.0));
   return MAX(0.3, MIN(1.0, 1.0 - fabs(tRInverse[line][nnext]) * 40.0));

	/* OK, we're not inside the racing line! */
#if 0
 // Check and see if we're significantly outside it and turning a corner, 
 // in which case we don't want to correct too much either.
 if ((tRInverse[line][Next] > 0.001 && car->_trkPos.toLeft > nlane2left + 2.0) ||
     (tRInverse[line][Next] < -0.001 && car->_trkPos.toLeft < nlane2left - 2.0))
  return MAX(0.2, MIN(1.0, 1.0 - fabs(tRInverse[line][Next]) * (car->_speed_x-40.0)*4));
 
 if ((tRInverse[line][nnext] > 0.001 && car->_trkPos.toLeft > nlane2left + 2.0) ||
     (tRInverse[line][nnext] < -0.001 && car->_trkPos.toLeft < nlane2left - 2.0))
  return MAX(0.2, MIN(1.0, 1.0 - fabs(tRInverse[line][nnext]) * (car->_speed_x-40.0)*2));
#endif
 // Check and see if we're outside it and turning into a corner,
 //  in which case we want to correct more to try and get closer to the apex.
 if ((tRInverse[line][Next] > 0.001 && tLane[Next] <= tLane[This] && car->_trkPos.toLeft > nlane2left + 2.0) ||
     (tRInverse[line][Next] < -0.001 && tLane[Next] >= tLane[This] && car->_trkPos.toLeft < nlane2left - 2.0))
  return MAX(1.0, MIN(1.5, 1.0 + fabs(tRInverse[line][Next])));

 return 1.0;
}

/**********************************************************/
double LRaceLine::getAvoidSpeed( float distance1, float distance2 )
{
 int i;
 //double speed1 = 1000.0, speed2 = 1000.0, speed3 = 1000.0;
 double speed1 = 1000.0, speed2 = 1000.0;

 int count = 0;
 for (i=Next; count < (int) (distance1/DivLength); i++)
 {
  count++;
  i = (i % Divs);
  speed1 = MIN(speed1, tSpeed[LINE_MID][i]);
 }

 count = 0;
 distance2 = (MIN(distance2, distance1 * 3) - distance1) / DivLength;
 for (; count < (int) distance2; i++)
 {
  count++;
  i = (i % Divs);
  speed2 = MIN(speed2, tSpeed[LINE_MID][i] + (double)count * 0.25);
 }

 return MIN(speed1, speed2);
} 

/********************************************************************/
/*                                                                  */
/*			STORE K1999 DATA                            */
/*		Raceline K1999 (aka Andrew's Locus)                 */
/*                                                                  */
/********************************************************************/
/* Set "savedata" to 1 in the xml file 
			   Not used in race, just for look at */

// saved in your Local directory .torcs/drivers/[module_name]/tracks/[track_name].data

//void LRaceLine::storeData( void **carParmHandle )
void LRaceLine::StoreData(tTrack* t)
{
  int rlLine = 0;

	
    	rlLine = GfParmGetNum( car->_carHandle, BT_SECT_PRIV, (char*)"savedata", (char*)NULL, 0);

 if (rlLine)
 {
	WriteLine(t);
	WriteTrack(t);
 }
  else
	return;

}

/**********************************************************/
void LRaceLine::WriteLine( tTrack* track )
{

 FILE *tfile = NULL;
 int i;
 int rl = LINE_RL;

  char filename[256];
	sprintf(filename, "%sdrivers/%s/tracks/%s_line.data", GetLocalDir(), myModuleName, track->internalname);

  //tfile = fopen( "/home/gamer/TRB/torcs_data/tuxy_dtm/tracks/wheel-2.learn", "w");

   tfile = fopen(filename, "w");

   if ( tfile == NULL )
      return;

  fprintf(tfile,"#==========================================================================================\n");
  fprintf(tfile,"# K1999 RaceLine data\n");
  fprintf(tfile,"# i tDistance[i] tLane[rl][i] tx[rl][i] ty[rl][i] tRInverse[rl][i] tSpeed[rl][i]\n");
  fprintf(tfile,"# Divs:%d\n",Divs);
    for ( i = 0; i < Divs ; i++ )
    {
      fprintf(tfile,"%d %f %f %f %f %f %f\n",
	      i,
	      tDistance[i],
	      tLane[rl][i],
	      tx[rl][i], ty[rl][i],
	      tRInverse[rl][i],
	      tSpeed[rl][i]
	      );
    }
  fprintf(tfile,"#End\n");
  fprintf(tfile,"#===============\n");
  fprintf(tfile,"\n");
  
  fflush(tfile);
  fclose(tfile);

}

/**********************************************************/

void LRaceLine::WriteTrack( tTrack* track )
{
  FILE *tfile = NULL;
 int i;

  char filename[256];
	sprintf(filename, "%sdrivers/%s/tracks/%s_track.data", GetLocalDir(), myModuleName, track->internalname);
  tfile = fopen(filename, "w");
  
  if ( tfile == NULL )
      return;

  fprintf(tfile,"#=============================================================================\n");
  fprintf(tfile,"# K1999 Track data\n");
  fprintf(tfile,"# i tDistance[i] txLeft[i] tyLeft[i] txRight[i] tyRight[i] tFriction[i]\n");
  fprintf(tfile,"# Divs:%d\n",Divs);

  for ( i = 0; i < Divs ; i++ )
    {
      fprintf(tfile,"%d %f %f %f %f %f %f\n",
	      i,
	      tDistance[i],
	      txLeft[i], tyLeft[i],
	      txRight[i], tyRight[i],
	      tFriction[i]
	      );
    }

  fprintf(tfile,"#Done\n");
  fprintf(tfile,"#=================================\n");
  fprintf(tfile,"\n");

  fflush(tfile);
  fclose(tfile);

}

/*==========================================================*/



